"""
Package Docuemntation
"""
# -*- coding: utf-8 -*-

from .insights import ModelXRay, explain_prediction_difference, explain_prediction_difference_xgboost
from .calibration import SplineCalibratedClassifierCV
from .splinecalib import SplineCalib
from .calibration_utils import prob_calibration_function, prob_calibration_function_multiclass
from .calibration_utils import compact_logit,inverse_compact_logit,plot_prob_calibration,plot_reliability_diagram
from .calibration_utils import get_stratified_foldnums,cv_predictions
from .modeling_utils import plot_pr_curve,plot_pr_curves,histogram_pair
from .shap_insights import consolidate_reason_scores, get_reason_codes, cv_column_shap, predict_reason_strings, predict_reasons_cv, get_reason_score_matrix
from . import metrics
from .CVModel import CVModel

__version__ = '0.1.1'



"""Calibration of predicted probabilities."""
import numpy as np
import sklearn
import warnings
from sklearn.base import BaseEstimator, ClassifierMixin, clone

try:
    from sklearn.model_selection import StratifiedKFold
except:
    from sklearn.cross_validation import StratifiedKFold

from .calibration_utils import prob_calibration_function, compact_logit


class SplineCalibratedClassifierCV(BaseEstimator, ClassifierMixin):
    """Probability calibration using cubic splines.

    With this class, the base_estimator is fit on each of the cross-validation
    training set folds in order to generate scores on the (cross-validated)
    test set folds.  The test set scores are accumulated into a final vector
    (the size of the full set) which is used to calibrate the answers.
    The model is then fit on the full data set.  The predict, and predict_proba
    methods are then updated to use the combination of the predictions from the 
    full model and the calibration function computed as above.

    Parameters
    ----------
    base_estimator : instance BaseEstimator
        The classifier whose output decision function needs to be calibrated
        to offer more accurate predict_proba outputs. If cv='prefit', the
        classifier must have been fit already on data.

    method : 'logistic' or 'ridge'
        The default is 'logistic', which is best if you plan to use log-loss as your
        performance metric.  This method is relatively robust and will typically do
        well on brier score as well.  The 'ridge' method calibrates using an L2 loss,
        and therefore should do better for brier score, but may do considerably worse
        on log-loss.

    cv : integer, cross-validation generator, iterable or "prefit", optional
        Determines the cross-validation splitting strategy.
        Possible inputs for cv are:

        - None, to use the default 5-fold cross-validation,
        - integer, to specify the number of folds.
        - 'prefit', if you wish to use the data only for calibration

        For integer/None inputs, if ``y`` is binary or multiclass,
        :class:`sklearn.model_selection.StratifiedKFold` is used. If ``y`` is
        neither binary nor multiclass, :class:`sklearn.model_selection.KFold`
        is used.

        Refer :ref:`User Guide <cross_validation>` for the various
        cross-validation strategies that can be used here.

        If "prefit" is passed, it is assumed that base_estimator has been
        fitted already and all data is used for calibration.

    Attributes
    ----------
    uncalibrated_classifier: this gives the uncalibrated version of the classifier, fit on the entire data set

    calib_func: this is the calibration function that has been learned from the cross-validation.  Applying this function
     to the results of the uncalibrated classifier (via model.predict_proba(X_test)[:,1]) gives the fully calibrated classifier

    References
    ----------
   """
    def __init__(self, base_estimator=None, method='logistic', cv=5, transform_type='none', cl_eps = .000001, **calib_kwargs):
        warn_msg = ('\nThis class is deprecated and will eventually be removed.' + 
                    '\nPlease use the SplineCalib class for calibration.')
        warnings.warn(warn_msg, FutureWarning)

        self.base_estimator = base_estimator
        self.uncalibrated_classifier = None
        self.calib_func = None
        self.method = method
        self.cv = cv
        self.cl_eps = cl_eps
        self.calib_kwargs = calib_kwargs
        self.fit_on_multiclass = False
        self.transform_type = transform_type
        self.pre_transform = lambda x: x
        if type(self.transform_type) == str:
            if self.transform_type == 'cl':
                self.pre_transform = lambda x: compact_logit(x, eps = self.cl_eps)
        if callable(self.transform_type):
            self.pre_transform = self.transform_type

    def fit(self, X, y, verbose=False):
        """Fit the calibrated model

        Parameters
        ----------
        X : array-like, shape (n_samples, n_features)
            Training data.

        y : array-like, shape (n_samples,)
            Target values.

        Returns
        -------
        self : object
            Returns an instance of self.
        """

        
        if len(np.unique(y)) > 2:
            self.fit_on_multiclass = True
            return self._fit_multiclass(X, y, verbose=verbose)

        self.fit_on_multiclass=False
        if ((type(self.cv)==str) and (self.cv=='prefit')):
            self.uncalibrated_classifier = self.base_estimator
            y_pred = self.uncalibrated_classifier.predict_proba(X)[:,1]

        else:
            y_pred = np.zeros(len(y))
            
            if sklearn.__version__ < '0.18':
                if type(self.cv)==int:
                    skf = StratifiedKFold(y, n_folds=self.cv,shuffle=True)
                else:
                    skf = self.cv
            else:
                if type(self.cv)==int:
                    skf = StratifiedKFold(n_splits=self.cv, shuffle=True).split(X, y)
                else:
                    skf = self.cv.split(X,y)
            for idx, (train_idx, test_idx) in enumerate(skf):
                if verbose:
                    print("training fold {} of {}".format(idx+1, self.cv))
                X_train = np.array(X)[train_idx,:]
                X_test = np.array(X)[test_idx,:]
                y_train = np.array(y)[train_idx]
                # We could also copy the model first and then fit it
                this_estimator = clone(self.base_estimator)
                this_estimator.fit(X_train,y_train)
                y_pred[test_idx] = this_estimator.predict_proba(X_test)[:,1]
            
            if verbose:
                print("Training Full Model")
            self.uncalibrated_classifier = clone(self.base_estimator)
            self.uncalibrated_classifier.fit(X, y)

        # calibrating function
        if verbose:
            print("Determining Calibration Function")
        if self.method=='logistic':
            self.calib_func = prob_calibration_function(y, self.pre_transform(y_pred), verbose=verbose, **self.calib_kwargs)
        if self.method=='ridge':
            self.calib_func = prob_calibration_function(y, self.pre_transform(y_pred), method='ridge', verbose=verbose, **self.calib_kwargs)
        # training full model

        return self

    def _fit_multiclass(self, X, y, verbose=False):
        """Fit the calibrated model in multiclass setting

        Parameters
        ----------
        X : array-like, shape (n_samples, n_features)
            Training data.

        y : array-like, shape (n_samples,)
            Target values.

        Returns
        -------
        self : object
            Returns an instance of self.
        """
        class_list = np.unique(y)
        num_classes = len(class_list)
        y_mod = np.zeros(len(y))
        for i in range(num_classes):
           y_mod[y==class_list[i]]=i

        y_mod = y_mod.astype(int)
        if ((type(self.cv)==str) and (self.cv=='prefit')):
            self.uncalibrated_classifier = self.base_estimator
            y_pred = self.uncalibrated_classifier.predict_proba(X)

        else:
            y_pred = np.zeros((len(y_mod),num_classes))
            if sklearn.__version__ < '0.18':
                skf = StratifiedKFold(y_mod, n_folds=self.cv,shuffle=True)
            else:
                skf = StratifiedKFold(n_splits=self.cv, shuffle=True).split(X, y)
            for idx, (train_idx, test_idx) in enumerate(skf):
                if verbose:
                    print("training fold {} of {}".format(idx+1, self.cv))
                X_train = np.array(X)[train_idx,:]
                X_test = np.array(X)[test_idx,:]
                y_train = np.array(y_mod)[train_idx]
                # We could also copy the model first and then fit it
                this_estimator = clone(self.base_estimator)
                this_estimator.fit(X_train,y_train)
                y_pred[test_idx,:] = this_estimator.predict_proba(X_test)
            
            if verbose:
                print("Training Full Model")
            self.uncalibrated_classifier = clone(self.base_estimator)
            self.uncalibrated_classifier.fit(X, y_mod)

        # calibrating function
        if verbose:
            print("Determining Calibration Function")
        if self.method=='logistic':
            self.calib_func, self.cf_list = prob_calibration_function_multiclass(y_mod, self.pre_transform(y_pred), verbose=verbose, **self.calib_kwargs)
        if self.method=='ridge':
            self.calib_func, self.cf_list = prob_calibration_function_multiclass(y_mod, self.pre_transform(y_pred), verbose=verbose, method='ridge', **self.calib_kwargs)
        # training full model

        return self


    def predict_proba(self, X):
        """Posterior probabilities of classification

        This function returns posterior probabilities of classification
        according to each class on an array of test vectors X.

        Parameters
        ----------
        X : array-like, shape (n_samples, n_features)
            The samples.

        Returns
        -------
        C : array, shape (n_samples, n_classes)
            The predicted probas.
        """
        # check_is_fitted(self, ["classes_", "calibrated_classifier"])
        if self.fit_on_multiclass:
            return self.calib_func(self.pre_transform(self.uncalibrated_classifier.predict_proba(X)))
        
        col_1 = self.calib_func(self.pre_transform(self.uncalibrated_classifier.predict_proba(X)[:,1]))
        col_0 = 1-col_1
        return np.vstack((col_0,col_1)).T
        
            

    def predict(self, X):
        """Predict the target of new samples. Can be different from the
        prediction of the uncalibrated classifier.

        Parameters
        ----------
        X : array-like, shape (n_samples, n_features)
            The samples.

        Returns
        -------
        C : array, shape (n_samples,)
            The predicted class.
        """
        # check_is_fitted(self, ["classes_", "calibrated_classifier"])
        return self.uncalibrated_classifier.classes_[np.argmax(self.predict_proba(X), axis=1)]

    def classes_(self):
        return self.uncalibrated_classifier.classes_



"""Calibration of predicted probabilities."""
import numpy as np
import sklearn
from sklearn.base import BaseEstimator, ClassifierMixin, clone

try:
    from sklearn.model_selection import StratifiedKFold
except:
    from sklearn.cross_validation import StratifiedKFold

from .calibration_utils import prob_calibration_function_multiclass


class SplineCalibratedClassifierMulticlassCV(BaseEstimator, ClassifierMixin):
    """Probability calibration using cubic splines.

    With this class, the base_estimator is fit on each of the cross-validation
    training set folds in order to generate scores on the (cross-validated)
    test set folds.  The test set scores are accumulated into a final vector
    (the size of the full set) which is used to calibrate the answers.
    The model is then fit on the full data set.  The predict, and predict_proba
    methods are then updated to use the combination of the predictions from the 
    full model and the calibration function computed as above.

    Parameters
    ----------
    base_estimator : instance BaseEstimator
        The classifier whose output decision function needs to be calibrated
        to offer more accurate predict_proba outputs. If cv='prefit', the
        classifier must have been fit already on data.

    method : 'logistic' or 'ridge'
        The default is 'logistic', which is best if you plan to use log-loss as your
        performance metric.  This method is relatively robust and will typically do
        well on brier score as well.  The 'ridge' method calibrates using an L2 loss,
        and therefore should do better for brier score, but may do considerably worse
        on log-loss.

    cv : integer, cross-validation generator, iterable or "prefit", optional
        Determines the cross-validation splitting strategy.
        Possible inputs for cv are:

        - None, to use the default 5-fold cross-validation,
        - integer, to specify the number of folds.
        - 'prefit', if you wish to use the data only for calibration

        For integer/None inputs, if ``y`` is binary or multiclass,
        :class:`sklearn.model_selection.StratifiedKFold` is used. If ``y`` is
        neither binary nor multiclass, :class:`sklearn.model_selection.KFold`
        is used.

        Refer :ref:`User Guide <cross_validation>` for the various
        cross-validation strategies that can be used here.

        If "prefit" is passed, it is assumed that base_estimator has been
        fitted already and all data is used for calibration.

    Attributes
    ----------
    uncalibrated_classifier: this gives the uncalibrated version of the classifier, fit on the entire data set

    calib_func: this is the calibration function that has been learned from the cross-validation.  Applying this function
     to the results of the uncalibrated classifier (via model.predict_proba(X_test)[:,1]) gives the fully calibrated classifier

    References
    ----------
   """
    def __init__(self, base_estimator=None, method='logistic', cv=5, **calib_kwargs):
        warn_msg = ('\nThis class is deprecated and will eventually be removed.' + 
                    '\nPlease use the SplineCalib class for calibration.')
        warnings.warn(warn_msg, FutureWarning)

        self.base_estimator = base_estimator
        self.uncalibrated_classifier = None
        self.calib_func = None
        self.method = method
        self.cv = cv
        self.calib_kwargs = calib_kwargs

    def fit(self, X, y, verbose=False):
        """Fit the calibrated model

        Parameters
        ----------
        X : array-like, shape (n_samples, n_features)
            Training data.

        y : array-like, shape (n_samples,)
            Target values.

        Returns
        -------
        self : object
            Returns an instance of self.
        """
        class_list = np.unique(y)
        num_classes = len(class_list)
        y_mod = np.zeros(len(y))

        for i in range(num_classes):
            y_mod[np.where(y==class_list[i])]=i

        y_mod = y_mod.astype(int)
        if ((type(self.cv)==str) and (self.cv=='prefit')):
            self.uncalibrated_classifier = self.base_estimator
            y_pred = self.uncalibrated_classifier.predict_proba(X)[:,1]

        else:
            y_pred = np.zeros((len(y_mod),num_classes))
            if sklearn.__version__ < '0.18':
                skf = StratifiedKFold(y_mod, n_folds=self.cv,shuffle=True)
            else:
                skf = StratifiedKFold(n_splits=self.cv, shuffle=True).split(X, y)
            for idx, (train_idx, test_idx) in enumerate(skf):
                if verbose:
                    print("training fold {} of {}".format(idx+1, self.cv))
                X_train = np.array(X)[train_idx,:]
                X_test = np.array(X)[test_idx,:]
                y_train = np.array(y_mod)[train_idx]
                # We could also copy the model first and then fit it
                this_estimator = clone(self.base_estimator)
                this_estimator.fit(X_train,y_train)
                y_pred[test_idx,:] = this_estimator.predict_proba(X_test)
            
            if verbose:
                print("Training Full Model")
            self.uncalibrated_classifier = clone(self.base_estimator)
            self.uncalibrated_classifier.fit(X, y_mod)

        # calibrating function
        if verbose:
            print("Determining Calibration Function")
        if self.method=='logistic':
            self.calib_func = prob_calibration_function_multiclass(y_mod, y_pred, verbose=verbose, **self.calib_kwargs)
        if self.method=='ridge':
            self.calib_func = prob_calibration_function_multiclass(y_mod, y_pred, verbose=verbose, method='ridge', **self.calib_kwargs)
        # training full model

        return self

    def predict_proba(self, X):
        """Posterior probabilities of classification

        This function returns posterior probabilities of classification
        according to each class on an array of test vectors X.

        Parameters
        ----------
        X : array-like, shape (n_samples, n_features)
            The samples.

        Returns
        -------
        C : array, shape (n_samples, n_classes)
            The predicted probas.
        """
        # check_is_fitted(self, ["classes_", "calibrated_classifier"])
        return self.calib_func(self.uncalibrated_classifier.predict_proba(X))


    def predict(self, X):
        """Predict the target of new samples. Can be different from the
        prediction of the uncalibrated classifier.

        Parameters
        ----------
        X : array-like, shape (n_samples, n_features)
            The samples.

        Returns
        -------
        C : array, shape (n_samples,)
            The predicted class.
        """
        # check_is_fitted(self, ["classes_", "calibrated_classifier"])
        return self.uncalibrated_classifier.classes_[np.argmax(self.predict_proba(X), axis=1)]

    def classes_(self):
        return self.uncalibrated_classifier.classes_



import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
from scipy.stats import beta
from sklearn.metrics import precision_recall_curve

def plot_pr_curve(truth_vec, score_vec, x_axis='precision', **kwargs):
    prec, rec, _ = precision_recall_curve(truth_vec,score_vec)
    if x_axis=='precision':
        plt.plot(prec[:-1], rec[:-1], **kwargs)
        plt.xlabel('Precision')
        plt.ylabel('Recall')
    else:
        plt.plot(rec[:-1], prec[:-1], **kwargs)
        plt.xlabel('Recall')
        plt.ylabel('Precision')
    plt.xlim([0,1])
    plt.ylim([0,1])

def plot_pr_curves(truth_vec_list, score_vec_list, x_axis='precision', **kwargs):
    for i in range(len(truth_vec_list)):
        prec, rec, _ = precision_recall_curve(truth_vec_list[i],score_vec_list[i])
        if x_axis=='precision':
            plt.plot(prec[:-1], rec[:-1], **kwargs)
            plt.xlabel('Precision')
            plt.ylabel('Recall')
        else:
            plt.plot(rec[:-1], prec[:-1], **kwargs)
            plt.xlabel('Recall')
            plt.ylabel('Precision')
        plt.xlim([0,1])
        plt.ylim([0,1])


def histogram_pair(value_vec, binary_vec, bins, smoothing_const=.01, prior_prob=.5, rel_risk=False, 
                    error_bar_alpha=.05,  figsize = (12,6), **kwargs):
    """This is a tool to explore the relationship between a numerical feature and a 1/0 binary outcome.

    Author: Brian Lucena

    It plots two histograms: one is of the values of the feature when the binary outcome is positive (1)
    and the other when it is negative (0).

    It then gives the marginal empirical probability of being a 1 given that the numerical feature
    is in a particular value range.

    In practice, it often takes some experimentation to find the appropriate bin endpoints for a
    particular feature.

    If the data contains 'NaN' values, it will also draw two small horizontal (dotted and dashed)
    lines, indicating the probabilities given NaN and not NaN respectively.
    """
    nan_mask = np.isnan(value_vec)
    num_nans = np.sum(nan_mask)
    if num_nans > 0:
        nan_binary_vec = binary_vec[nan_mask]
        binary_vec = binary_vec[~nan_mask]
        value_vec = value_vec[~nan_mask]
        nan_avg_value = np.mean(nan_binary_vec)
        reg_avg_value = np.mean(binary_vec)
    # digitized_value_vec = np.digitize(value_vec, bins)
    # x_pts_to_graph = np.array([np.mean(value_vec[digitized_value_vec==i]) for i in np.unique(digitized_value_vec)])
    # print(x_pts_to_graph)
    out0 = plt.hist(value_vec[binary_vec == 0], bins=bins, **kwargs)
    out1 = plt.hist(value_vec[binary_vec == 1], bins=bins, **kwargs)
    plt.close()
    plt.figure(figsize=figsize)
    plt.subplot(2, 1, 1)
    plt.hist((value_vec[binary_vec == 0],value_vec[binary_vec == 1]), stacked=True, bins=bins, **kwargs)
    bin_leftpts = (out1[1])[:-1]
    bin_rightpts = (out1[1])[1:]
    default_bin_centers = (bin_leftpts + bin_rightpts) / 2
    digitized_value_vec = np.digitize(value_vec, bins)
    bin_centers = np.array([np.mean(value_vec[digitized_value_vec==i]) if i in np.unique(digitized_value_vec) else default_bin_centers[i-1] for i in np.arange(len(bins)-1)+1])
    prob_numer = out1[0]
    prob_denom = out1[0] + out0[0]
    smoothing_const = .001
    probs = (prob_numer + prior_prob * smoothing_const) / (prob_denom + smoothing_const)
    # print(bin_centers)
    # print(probs)
    plt.subplot(2, 1, 2)
    if rel_risk:
        plt.plot(bin_centers, np.log10(probs / prior_prob))
        # plt.errorbar(bin_centers, probs, yerr=1.96 * probs * (1 - probs) / np.sqrt(prob_denom), capsize=3)
        plt.xlim(bin_leftpts[0], bin_rightpts[-1])
    else:
        plt.plot(bin_centers[:len(probs)], probs)
        plt.xlim(bin_leftpts[0], bin_rightpts[-1])
        yerr_mat_temp = beta.interval(1-error_bar_alpha,out1[0]+1,out0[0]+1)
        yerr_mat = np.vstack((yerr_mat_temp[0],yerr_mat_temp[1])) - probs
        yerr_mat[0,:] = -yerr_mat[0,:]
        plt.errorbar(bin_centers[:len(probs)], probs, yerr=yerr_mat, capsize=5)
        plt.xlim(bin_leftpts[0], bin_rightpts[-1])
        if num_nans > 0:
            plt.hlines(y=nan_avg_value, xmin=bin_leftpts[0], xmax=bin_leftpts[1], linestyle='dotted')
            plt.hlines(y=reg_avg_value, xmin=bin_leftpts[0], xmax=bin_leftpts[1], linestyle='dashed')
    return {'bin_centers': bin_centers, 'probs': probs, 'prob_numer': prob_numer, 'prob_denom': prob_denom}







import math
import warnings
import numpy as np
import pandas as pd
from .utils import _gca, is_classifier, is_regressor


class ModelXRay(object):
    """This class executes a model over a broad range of modified data points to analyze aspects of its performance.

    For each point in the data set, and for every feature involved of the prediction of the model, a new set of data
    points is created where the chosen feature is varied across its (empirical) range.  These modified data points are
    fed into the model to get a set of model predictions for each feature-data point combination.

    It is desirable that the "data" object passed in be relatively large in size, since the algorithm will make
    some heuristic choices based on the ranges of values it sees.  We suggest using at least 100 data points and preferably
    more than 500.

    It returns a results object, which can then be passed to functions such as feature_effect_summary and
    feature_dependence_plots to gain insight on the how the various features affect the target.  The results
    object can also be used directly by a user who wants to operate at a low-level.

       Parameters
        ----------


        model : A model object from sklearn or similar styled objects.  The `predict` method will be used if it is
            a regression model, while `predict_proba` will be used if it is a (binary) classification model.  Multi-class
            classifiers are not supported at this time.

        data : A DataFrame possessing the sameucture that the model would take as an argument.  These methods are designed
            to be used on "test" data (i.e. data that was not used in the training of the model).  However, there is nothing
            structural to prevent it from being used on training data, and there may be some insight gained by doing so.

        columns : a specific subset of columns to be used.  Default is None, which means to use all available columns in *data*

        resolution : how many different "grid points" to use for each feature.  The algorithm will use only the unique values
        it sees in *data* if there are fewer than *resolution* unique values.  Otherwise it will use *resolution* linearly spaced
        values ranging from the min observed value to the max observed value.
    """

    def __init__(self, model, data, columns=None, resolution=100, normalize_loc=None, pred_col_name = None, pred_col_index=1):

        self.model = model
        self.data = data
        self.pred_col_index = pred_col_index
        if type(data) == pd.DataFrame:
            if (pred_col_name != None) and (is_classifier(self.model)):
                self.pred_col_index = np.where(self.model.classes_ == pred_col_name)[0][0]
            self.pred_col_name = data.columns[self.pred_col_index]
            self.data_values = data.values


        else:
            self.data_values = data

        self.columns = columns
        self.results = self._model_xray(columns, resolution, normalize_loc)


    def _get_data_rows(self, row_nums):
        if type(self.data) == pd.DataFrame:
            return self.data.iloc[row_nums]
        else:
            return self.data[row_nums, :]


    def _get_predictions(self, rows):
        # Catch deprecated warnings from Predict call
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")

            if is_classifier(self.model):
                y_pred = self.model.predict_proba(rows)[:,self.pred_col_index]
            else:
                #print('off')
                y_pred = self.model.predict(rows)
        return y_pred


    def gen_model_pred(self, row, col_idx, values):
        rows = []
        for val in values:
            sim_row = row.copy()
            sim_row[col_idx] = val
            rows.append(sim_row)
        # If the row is a Series, make it into a DF
        if type(rows[0]) == pd.Series:
            rows = pd.DataFrame(rows)
        y_pred = self._get_predictions(rows)
        return y_pred

    def _model_xray(self, columns, resolution, normalize_loc):
        '''This function executes a model over a broad range of conditions to analyze aspects of its performance.

        For each point in the data set, and for every feature involved of the prediction of the model, a new set of data
        points is created where the chosen feature is varied across its (empirical) range.  These modified data points are
        fed into the model to get a set of model predictions for each feature-data point combination.

        It is desirable that the "data" object passed in be relatively large in size, since the algorithm will make
        some heuristic choices based on the ranges of values it sees.  We suggest using at least 100 data points and preferably
        more than 500.

        It returns a results object, which can then be passed to functions such as feature_effect_summary and
        feature_dependence_plots to gain insight on the how the various features affect the target.  The results
        object can also be used directly by a user who wants to operate at a low-level.

            Parameters
            ----------

            model : A model object from sklearn or similar styled objects.  The `predict` method will be used if it is
                a regression model, while `predict_proba` will be used if it is a (binary) classification model.  Multi-class
                classifiers are not supported at this time.

            data : A DataFrame possessing the same structure that the model would take as an argument.  These methods are designed
                to be used on "test" data (i.e. data that was not used in the training of the model).  However, there is nothing
                structural to prevent it from being used on training data, and there may be some insight gained by doing so.

            columns : a specific subset of columns to be used.  Default is None, which means to use all available columns in *data*

            resolution : how many different "grid points" to use for each feature.  The algorithm will use only the unique values
            it sees in *data* if there are fewer than *resolution* unique values.  Otherwise it will use *resolution* linearly spaced
            values ranging from the min observed value to the max observed value.

            Returns
            -------

            results : The "results" object is a dictionary where the keys are the feature names and the values are a 2-tuple.  This
                object is intended primarily to be passed to other functions to interact with and display the data.  However, advanced
                users may wish to understand and/or use the object directly.

                The first element in the tuple is the set of different feature values that were substituted in for each data point.  The
                second element in the tuple is matrix where the number of rows is the number of data points and the number of columns
                is the number of different feature values.  The (i,j)th element of the matrix is the result of the model prediction when
                data point i has the feature in question set to jth value.
            '''
        ## Convert Pandas DataFrame to nparray explicitly to make life easier
        #print('hello!!!')


        ## Determine the range of values to plot for the chosen column
        if columns is None:
            if type(self.data) == pd.DataFrame:
                columns = self.data.columns
            if type(self.data)==np.ndarray:
                columns = range(len(self.data[0]))  # Assuming a 2-D Dataset
        else:
            # Verify that columns is an iterable
            try:
                iterator = iter(columns)
            except TypeError:
                # not iterable
                columns = [columns]
            else:
                # iterable
                pass

        # Build Column Index
        column_nums = []
        if type(self.data) == pd.DataFrame:
            for column in columns:
                try:
                    column_nums.append(self.data.columns.get_loc(column))
                except KeyError:
                    ## TODO
                    pass
        else:
            # Column Index and Column Names are the same
            if type(columns[0]) == int:
                column_nums = columns
            else:
                column_nums = range(len(columns))

        # Use the Numpy array of data values to ease indexing by col. numbers
        results = {}
        num_pts = len(self.data_values)
        for column_num, column_name in zip(column_nums, columns):
            if (len(np.unique(self.data_values[:,column_num])) > resolution):
                col_values = np.linspace(np.nanmin(self.data_values[:,column_num]),
                    np.nanmax(self.data_values[:,column_num]),resolution)
            else:
                col_values = np.sort(np.unique(self.data_values[:,column_num]))
            ## Define the empty data structure to output
            out_matrix = np.zeros([num_pts,len(col_values)])

            ## Generate predictions
            if type(self.data) == pd.DataFrame:
                rows = self.data.iterrows()
            else:
                rows = enumerate(self.data)
            for loop_idx, (row_idx, row) in enumerate(rows):
                y_pred = self.gen_model_pred(row, column_num, col_values)
                if normalize_loc=='start':
                    y_pred = y_pred - y_pred[0]
                if normalize_loc=='end':
                    y_pred = y_pred - y_pred[-1]
                if (type(normalize_loc)==int and normalize_loc>=0 and normalize_loc<resolution):
                    y_pred = y_pred - y_pred[normalize_loc]
                out_matrix[loop_idx,:] = y_pred
            results[column_name] = (col_values, out_matrix)
        return results


    def feature_effect_summary(self, kind="boxh", num_features=20, y_scaling = 'none', ax=None):
        '''This function plots a comparison of the effects of different features in a complex predictive model.

        In more complicated predictive models, the effect of an individual feature can be highly dependent on the values
        of the other features.  It could be that a feature has a large effect in one context but a negligible effect in another.
        This visualization attempts to shed light on the range of possibilities of the effect of a feature, by giving a boxplot
        showing the range of possibilities of the effect of a feature.

        The features are ranked by their "median" effect across a range of data points, where the "effect"
        is measured by the "peak to trough" distance that occurs as that feature varies across its possible range.

            Parameters
            ----------

            results : This is a results object from a call to model_xray.  Ideally, the model_xray was given a reasonably large amount of data
                so that we can empirically see a broad range of possibilities.

            kind : Currently only 'boxh' (horizontal boxplot) is supported

            ax : If desired, a particular axis on which to generate the plot can be passed to the function

            num_features : This specifies the maximum number of features to include in the boxplot.  The function chooses the most significant
                features as measured by the median peak-to-trough effect size.

            Returns
            -------
        '''
        ## Convert Pandas DataFrame to nparray explicitly to make life easier

        columns = list(self.results.keys())
        logit_func = lambda x: np.log(x/(1-x))
        logit10_func = lambda x: np.log10(x/(1-x))
        logit2_func = lambda x: np.log2(x/(1-x))
        if y_scaling=='logit':
            result_data = [importance_distribution_of_variable(logit_func(self.results[col_name][1])) for col_name in columns]
        elif y_scaling=='logit10':
            result_data = [importance_distribution_of_variable(logit10_func(self.results[col_name][1])) for col_name in columns]
        elif y_scaling=='logit2':
            result_data = [importance_distribution_of_variable(logit2_func(self.results[col_name][1])) for col_name in columns]
        else:
            result_data = [importance_distribution_of_variable(self.results[col_name][1]) for col_name in columns]
        sortind = np.argsort([np.median(d) for d in result_data])
        if num_features and num_features > 0:
            num_features = min(num_features, len(columns))
        else:
            num_features = len(columns)
        plot_data = [result_data[idx] for idx in sortind][-num_features:]

        if ax is None:
            ax = _gca()
            fig = ax.get_figure()
            fig.set_figwidth(10)
            fig.set_figheight(max(6, int(math.ceil(num_features*0.5))))
        ax.boxplot(plot_data, notch=0, sym='+', vert=0, whis=1.5)
        ax.set_yticklabels([columns[idx] for idx in sortind][-num_features:]);


    def feature_dependence_plots(self, y_scaling='none', show_base_points=True, pts_selected='sample',
        columns = None, num_pts=5, figsize=None):
        '''This function visualizes the effect of a single variable in models with complicated dependencies.
        Given a dataset, it will select points in that dataset, and then change the select column across
        different values to view the effect of the model prediction given that variable. These have been called
        Individual Conditional Expectation plots (or ICE-plots), see Goldstein, Kapelner, Bleich,
        Pitkin. Peeking Inside the Black Box: Visualizing Statistical Learning With Plots of Individual
        Conditional Expectation. Journal of Computational and Graphical Statistics (March 2014)
        '''

        import matplotlib.pyplot as plt

        if columns == None:
            columns = sorted(list(self.results.keys()))
        num_rows = len(self.results[columns[0]][1])  # Get number of sample rows
        if (type(pts_selected)==str and pts_selected=='sample'):
            row_indexes = np.random.choice(np.arange(num_rows), num_pts)
        else:
            row_indexes = pts_selected

        if show_base_points:
            base_rows = self._get_data_rows(row_indexes)
            y_base_points = self._get_predictions(base_rows)
            if y_scaling=='logit':
                y_base_points = np.log(y_base_points/(1-y_base_points))
            if y_scaling=='logit10':
                y_base_points = np.log10(y_base_points/(1-y_base_points))
            if y_scaling=='logit2':
                y_base_points = np.log2(y_base_points/(1-y_base_points))
        else:
            y_base_points = None

        n_cols = min(3, len(columns))
        n_rows = int(math.ceil(len(columns) / n_cols))
        figsize = (n_cols * 4, n_rows * 4)
        fig, axes = plt.subplots(n_rows, n_cols, figsize=figsize)
        for col_name, ax in zip(columns, axes.flatten()):
            x = self.results[col_name][0]
            y_values = self.results[col_name][1][row_indexes]
            y_plot = y_values
            if y_scaling=='logit':
                y_plot = np.log(y_values/(1-y_values))
            if y_scaling=='logit10':
                y_plot = np.log10(y_values/(1-y_values))
            if y_scaling=='logit2':
                y_plot = np.log2(y_values/(1-y_values))
            for y in y_plot:
                ax.plot(x, y)
            # Plot Base Points
            if y_base_points is not None:
                ax.scatter(base_rows[col_name], y_base_points)
            ax.set_title(col_name[:30])
        plt.tight_layout()
        return row_indexes


    def explain_prediction_difference(self, index_1, index_2, tol=.03, verbose=True, decimals=4):
        '''Given the indices of two points in the "xray"-ed data set, this function gives an explanation
        of the factors contributing to the difference in the predictions.

        Starting with the first point given, the considers changing each feature from its current value to that
        possessed by the second point.  The function evaluates the target in both scenarios and determines the
        feature value change that creates the biggest (absolute) change in the target.  This change is selected
        and the current point becomes the new point with the new feature value.  This is repeated until the new
        target value is within a factor of 1+tol of the second point.
        '''
        data_row_1 = self._get_data_rows(index_1)
        data_row_2 = self._get_data_rows(index_2)
        return explain_prediction_difference(self.model, data_row_1, data_row_2, tol, verbose, decimals, self.pred_col_index)


def importance_distribution_of_variable(model_result_array):
    max_result_vec = np.array(list(map(np.max,model_result_array)))
    min_result_vec = np.array(list(map(np.min,model_result_array)))
    return max_result_vec - min_result_vec


def explain_prediction_difference(model, data_row_1, data_row_2, tol=.03, verbose=True, decimals = 4, pred_col_index=1):
    '''Given a model and two single row data frames, this function gives an explanation
        of the factors contributing to the difference in the predictions.

        Starting with the first point given, the considers changing each feature from its current value to that
        possessed by the second point.  The function evaluates the target in both scenarios and determines the
        feature value change that creates the biggest (absolute) change in the target.  This change is selected
        and the current point becomes the new point with the new feature value.  This is repeated until the new
        target value is within a factor of 1+tol of the second point.
        '''
    column_names = data_row_1.index
    num_columns = len(column_names)

    dr_1 = data_row_1.values.reshape(1,-1)
    dr_2 = data_row_2.values.reshape(1,-1)
    column_list = list(range(num_columns))
    curr_pt = np.copy(dr_1)
    if is_classifier(model):
        val1 = model.predict_proba(dr_1)[0,pred_col_index]
        val2 = model.predict_proba(dr_2)[0,pred_col_index]
    else:
        val1 = model.predict(dr_1)[0]
        val2 = model.predict(dr_2)[0]
    if verbose:
        print(val1, val2)
        print('Your initial point has a target value of {}'.format(np.round(val1,decimals=decimals)))
        print('Your final point has a target value of {}'.format(np.round(val2,decimals=decimals)))
    pt_list = [dr_1]
    val_list = [val1]
    curr_val = val1
    final_val = val2
    feat_list =[]
    move_list = []
    feat_val_change_list = []
    #for num_steps in range(4):
    while (((curr_val/final_val) >(1+tol)) or ((curr_val/final_val) <(1-tol))):
        biggest_move = 0
        best_column = -1
        best_val = curr_val
        for i in column_list:
            test_pt = np.copy(curr_pt)
            prev_feat_val = test_pt[0,i]
            subst_val = dr_2[0,i]
            test_pt[0,i] = subst_val
            if is_classifier(model):
                test_val = model.predict_proba(test_pt)[0,pred_col_index]
            else:
                test_val = model.predict(test_pt)[0]
            move_size = (test_val - curr_val)
            if(np.abs(move_size)>=np.abs(biggest_move)):
                biggest_move = move_size
                best_column = i
                best_val = test_val
                old_feat_val = prev_feat_val
                new_feat_val = subst_val
        subst_val = dr_2[0,best_column]
        curr_pt[0,best_column] = subst_val
        val_list.append(best_val)
        curr_val = best_val
        if verbose:
            print('Changing {} from {} to {}'.format(column_names[best_column],np.round(old_feat_val,decimals=decimals),np.round(new_feat_val,decimals=decimals)))
            print('\t\tchanges your target by {} to {}'.format(np.round(biggest_move,decimals=decimals), np.round(best_val,decimals=decimals)))
            print('----------')
            if not (((curr_val/final_val) >(1+tol)) or ((curr_val/final_val) <(1-tol))):
                print('Tolerance of {} reached'.format(tol))
                print('Current value of {} is within {}% of {}'.format(np.round(curr_val,decimals=decimals),(100*tol),np.round(final_val,decimals=decimals)))
        feat_list.append(column_names[best_column])
        column_list.remove(best_column)
        move_list.append(biggest_move)
        feat_val_change_list.append((old_feat_val, new_feat_val))
    return feat_list, feat_val_change_list, move_list, val_list


def explain_prediction_difference_xgboost(model, data_row_1, data_row_2, tol=.03, verbose=True, decimals = 4, pred_col_index=1):
    '''Given a model and two single row data frames, this function gives an explanation
        of the factors contributing to the difference in the predictions.

        Starting with the first point given, the considers changing each feature from its current value to that
        possessed by the second point.  The function evaluates the target in both scenarios and determines the
        feature value change that creates the biggest (absolute) change in the target.  This change is selected
        and the current point becomes the new point with the new feature value.  This is repeated until the new
        target value is within a factor of 1+tol of the second point.
        '''
    column_names = data_row_1.columns
    num_columns = len(column_names)

    #dr_1 = data_row_1.values.reshape(1,-1)
    #dr_2 = data_row_2.values.reshape(1,-1)
    dr_1 = data_row_1
    dr_2 = data_row_2
    column_list = list(range(num_columns))
    curr_pt = (dr_1).copy()
    if is_classifier(model):
        val1 = model.predict_proba(dr_1)[0,pred_col_index]
        val2 = model.predict_proba(dr_2)[0,pred_col_index]
    else:
        val1 = model.predict(dr_1)[0]
        val2 = model.predict(dr_2)[0]
    if verbose:
        print(val1, val2)
        print('Your initial point has a target value of {}'.format(np.round(val1,decimals=decimals)))
        print('Your final point has a target value of {}'.format(np.round(val2,decimals=decimals)))
    pt_list = [dr_1]
    val_list = [val1]
    curr_val = val1
    final_val = val2
    feat_list =[]
    move_list = []
    feat_val_change_list = []
    #for num_steps in range(4):
    while (((curr_val/final_val) >(1+tol)) or ((curr_val/final_val) <(1-tol))):
        biggest_move = 0
        best_column = -1
        best_val = curr_val
        for i in column_list:
            test_pt = (curr_pt).copy()
            prev_feat_val = test_pt.iloc[0,i]
            subst_val = dr_2.iloc[0,i]
            test_pt.iloc[0,i] = subst_val
            if is_classifier(model):
                test_val = model.predict_proba(test_pt)[0,pred_col_index]
            else:
                test_val = model.predict(test_pt)[0]
            move_size = (test_val - curr_val)
            if(np.abs(move_size)>=np.abs(biggest_move)):
                biggest_move = move_size
                best_column = i
                best_val = test_val
                old_feat_val = prev_feat_val
                new_feat_val = subst_val
        subst_val = dr_2.iloc[0,best_column]
        curr_pt.iloc[0,best_column] = subst_val
        val_list.append(best_val)
        curr_val = best_val
        if verbose:
            print('Changing {} from {} to {}'.format(column_names[best_column],np.round(old_feat_val,decimals=decimals),np.round(new_feat_val,decimals=decimals)))
            print('\t\tchanges your target by {} to {}'.format(np.round(biggest_move,decimals=decimals), np.round(best_val,decimals=decimals)))
            print('----------')
            if not (((curr_val/final_val) >(1+tol)) or ((curr_val/final_val) <(1-tol))):
                print('Tolerance of {} reached'.format(tol))
                print('Current value of {} is within {}% of {}'.format(np.round(curr_val,decimals=decimals),(100*tol),np.round(final_val,decimals=decimals)))
        feat_list.append(column_names[best_column])
        column_list.remove(best_column)
        move_list.append(biggest_move)
        feat_val_change_list.append((old_feat_val, new_feat_val))
    return feat_list, feat_val_change_list, move_list, val_list







import numpy as np
import pandas as pd
from sklearn.model_selection import StratifiedKFold

def cv_predict_proba(X, y,estimator, cv):
    
    ## Convert from Pandas to Numpy if necessary
    if (type(X)==pd.DataFrame)  or (type(X)==pd.Series):
        X = X.values
    if (type(y)==pd.DataFrame)  or (type(y)==pd.Series):
        y = y.values

    num_classes = len(np.unique(y))
    out_vec=np.zeros((len(y),num_classes))

    #Main loop to do cross-validated predict proba and construct output matrix
    for tr, te in cv.split(X,y):
        estimator.fit(X[tr],y[tr])
        out_vals = estimator.predict_proba(X[te])
        out_vec[te,:] = out_vals
    return out_vec


def cv_score(X, y, estimator, cv, score_fn):
    return(score_fn(y,cv_predict_proba(X,y,estimator,cv)))

def _get_param_settings_from_grid(param_grid):
    num_settings = np.prod([len(i) for i in param_grid.values()])
    pg_tuple = tuple(param_grid.items())
    param_names = [k[0] for k in pg_tuple]
    param_lists = [k[1] for k in pg_tuple]
    param_list_lengths = [len(k) for k in param_lists]
    param_dict_list = []
    for i in range(num_settings):
        indices = _int_to_indices(i,param_list_lengths)
        curr_param_dict = {}
        for k in range(len(param_names)):
            curr_param_dict[param_names[k]]=param_lists[k][indices[k]]
        param_dict_list.append(curr_param_dict)
    return param_dict_list    
    
def _int_to_indices(j,lengths):
    out_list = []
    for i in range(len(lengths)):
        curr_ind = j % lengths[i]
        out_list.append(curr_ind)
        j = j//lengths[i]
    return(out_list)

def grid_search(X,y, model, param_grid, score_fn, verbose=True):
    param_arg_list = _get_param_settings_from_grid(param_grid)
    num_settings = len(param_arg_list)
    param_list_scores = np.zeros(num_settings)
    skf = StratifiedKFold(5, shuffle=True, random_state=42)
    for i in range(num_settings):
        curr_param_dict = param_arg_list[i]
        if verbose:
            print(curr_param_dict)
        model.set_params(**curr_param_dict)
        curr_score=cv_score(X,y,model,skf,score_fn)
        param_list_scores[i]=curr_score
        if verbose:
            print(curr_param_dict,curr_score)

    return(list(zip(param_arg_list,param_list_scores)))




#!/usr/bin/env python
# -*- coding: utf-8 -*-

from setuptools import setup

with open('README.rst') as readme_file:
    readme = readme_file.read()

requirements = [
    "pandas>=0.23",
    "numpy>=1.16.0",
    "scikit-learn>=0.18.0",
    "scipy>=1.0"
]

setup(
    name='ml_insights',
    version='0.1.1',
    description="Package to understand ML Models",
    long_description=readme,
    author="Brian Lucena / Ramesh Sampath",
    author_email='brianlucena@gmail.com',
    url='http://ml-insights.readthedocs.io/en/latest/',
    packages=[
        'ml_insights',
    ],
    package_dir={'ml_insights':
                 'ml_insights'},
    include_package_data=True,
    install_requires=requirements,
    license="MIT license",
    zip_safe=False,
    keywords='ml_insights',
    classifiers=[
        'Development Status :: 2 - Pre-Alpha',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Natural Language :: English',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8'
    ]
)



import numpy as np
import pandas as pd

try:
    import xgboost as xgb

except ImportError:
    xgb_installed = False

def consolidate_reason_scores(df_ind_expl, dict_map):
    reason_list = dict_map.keys()
    df_rsn = pd.DataFrame(columns = reason_list)
    for reason in reason_list:
        df_rsn[reason] = np.sum(df_ind_expl.loc[:,dict_map[reason]], axis=1)
    return df_rsn 

def get_reason_codes(df_rsn, thresh, direction='greater', delimiter=';'):
    nr, nc = df_rsn.shape
    argsort_mat = np.argsort(-df_rsn.values)
    if (direction=='lesser'):
        num_exceeding_thresh_vec = np.sum(df_rsn.values<=thresh, axis=1)
    else:
        num_exceeding_thresh_vec = np.sum(df_rsn.values>=thresh, axis=1)
    reason_mat = np.array([df_rsn.columns[i] for row in argsort_mat for i in row ]).reshape(nr,nc)
    reason_vec = np.array([delimiter.join(list(reason_mat[j][:num_exceeding_thresh_vec[j]])) for j in range(nr)])
    return reason_vec

def cv_column_shap(xgbcv, X_pr, fn):
    results = np.zeros((X_pr.shape[0], xgbcv.num_features+1))
    fold_set = np.unique(fn)
    for fold in fold_set:
        X_te = xgb.DMatrix(X_pr[fn == fold].values)
        fold_results = xgbcv.model_dict[fold].get_booster().predict(X_te, pred_contribs=True, validate_features=False)
        results[fn==fold] = fold_results
    return results
    
def predict_reasons_cv(xgbcv, X_pr, fn, reason_map, thresh, delimiter=';'):
	shap_val_mat = cv_column_shap(xgbcv, X_pr, fn)
	df_shap_val = pd.DataFrame(shap_val_mat[:,:-1], columns = X_pr.columns)
	df_reason_scores = consolidate_reason_scores(df_shap_val,reason_map)
	reason_list_vec = get_reason_codes(df_reason_scores, thresh, delimiter=delimiter)
	return(reason_list_vec)

def predict_reason_strings(xgbmodel, X_pr, reason_map, thresh, delimiter=';', direction='greater'):
    X_pr_dmat = xgb.DMatrix(X_pr)
    shap_val_mat = xgbmodel.get_booster().predict(X_pr_dmat, pred_contribs=True, validate_features=False)
    df_shap_val = pd.DataFrame(shap_val_mat[:,:-1], columns = X_pr.columns)
    df_reason_scores = consolidate_reason_scores(df_shap_val,reason_map)
    reason_list_vec = get_reason_codes(df_reason_scores, thresh, direction=direction, delimiter=delimiter)
    return(reason_list_vec)

def get_reason_score_matrix(xgbmodel, X_pr, validate=False):
    if (type(X_pr)==pd.DataFrame):
        X_test_dmat = xgb.DMatrix(X_pr)
        reason_list = list(X_pr.columns) + ['Intercept']
        reas_mat = xgbmodel.get_booster().predict(X_test_dmat, pred_contribs=True, validate_features=validate)
    else:
        reason_list = ['f'+str(i) for i in range(X_pr.shape[1])] + ['Intercept']
        X_test_dmat = xgb.DMatrix(X_pr, feature_names = reason_list[:-1])
        reas_mat = xgbmodel.get_booster().predict(X_test_dmat, pred_contribs=True, validate_features=validate)
    return(pd.DataFrame(reas_mat, columns=reason_list))

# def augment_tree(tree_dict):
#     if 'leaf' in tree_dict.keys():
#         value = tree_dict['leaf']
#         tree_dict['value_at_node'] = value
#         return value
#     else:
#         a0 = tree_dict['children'][0]['cover']
#         a1 = tree_dict['children'][1]['cover']
#         value = (a0 * augment_tree(tree_dict['children'][0]) + a1 * augment_tree(tree_dict['children'][1]))/(a0 + a1)
        
#         tree_dict['value_at_node'] = value
#         return value

